import json
import re

from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pydantic import BaseModel
from typing import List, Dict
import logging
import time


logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)


class Tweet(BaseModel):
    text: str
    timestamp: str


class TwitterContent(BaseModel):
    tweets: List[Tweet] = []
    page_source: str = ""


class TwitterScraper:
    def __init__(self, chromedriver_path="/usr/local/bin/chromedriver", headless=True):
        self.chromedriver_path = chromedriver_path
        self.headless = headless
        self.driver = None
        self.username = None
        self.password = None

    def _setup_driver(self):
        chrome_options = Options()
        if self.headless:
            chrome_options.add_argument("--headless")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--no-sandbox")
        chrome_options.add_argument("--disable-dev-shm-usage")
        chrome_options.add_argument("--start-maximized")
        service = Service(self.chromedriver_path)
        return webdriver.Chrome(service=service, options=chrome_options)

    def login_to_twitter(self, username, password):
        self.driver = self._setup_driver()
        self.driver.get("https://x.com/login")
        self.username = username
        self.password = password

        # Wait for the username field and enter the username
        username_field = WebDriverWait(self.driver, 10).until(
            EC.presence_of_element_located((By.NAME, "text"))
        )
        username_field.send_keys(username)
        self.driver.find_element(By.XPATH, "//span[text()='Next']").click()

        # Wait for the password field and enter the password
        password_field = WebDriverWait(self.driver, 10).until(
            EC.presence_of_element_located((By.NAME, "password"))
        )
        password_field.send_keys(password)
        self.driver.find_element(By.XPATH, "//span[text()='Log in']").click()

        # Wait for the home page to load
        WebDriverWait(self.driver, 10).until(
            EC.presence_of_element_located((By.XPATH, "//div[@data-testid='primaryColumn']"))
        )

    def get_home_page_content(self,num_tweets=2):
        if not self.driver:
            raise Exception("You must login first using login_to_twitter method")

        # Scroll down a few times to load more tweets
        page_content = ""

        # Scroll down a few times to load more tweets
        for _ in range(num_tweets):
            page_content += self.driver.page_source
            self.driver.execute_script("window.scrollTo(0, document.body.scrollHeight);")
            WebDriverWait(self.driver, 10).until(
                EC.presence_of_element_located((By.XPATH, "//article[@data-testid='tweet']"))
            )
        with open("home_page_tweets.txt","w") as f:
            f.write(f'{page_content}')

        tweets= self.get_tweets_from_text(self.username,page_content)

        return TwitterContent(
            tweets=tweets,
            page_source=page_content
        )

    def get_tweets_from_text(self,username,text):
        spans = re.findall('<span(.*?)>\s*(.*?)\s*</span>', text, re.DOTALL)
        spans = [_[1] for _ in spans]
        tweets = []
        isp = []
        start = False
        for span in spans:
            if span and span[0] == '@':
                start = True
                isp = []
            if start:
                if 'span' in span:
                    start = False
                    tweets.append("".join(isp))
                else:
                    if username not in span:
                        isp.append(span)
        return [Tweet(text = t,timestamp = "") for t in tweets]

    def get_user_tweets(self, username: str, num_tweets: int = 2):
        if not self.driver:
            raise Exception("You must login first using login_to_twitter method")

        self.driver.get(f"https://x.com/{username}")



        # # Wait for tweets to load
        # WebDriverWait(self.driver, 10).until(
        #     EC.presence_of_element_located((By.XPATH, f'{username}'))
        # )
        time.sleep(1)

        page_content = ""

        # Scroll down a few times to load more tweets
        for _ in range(num_tweets):
            page_content += self.driver.page_source
            self.driver.execute_script("window.scrollTo(0, document.body.scrollHeight);")
            WebDriverWait(self.driver, 10).until(
                EC.presence_of_element_located((By.XPATH, "//article[@data-testid='tweet']"))
            )

        tweets = self.get_tweets_from_text(username = username,
                                  text = page_content)

        return TwitterContent(
            tweets=tweets,
            page_source=self.driver.page_source
        )

    def parse_saved_html(self, content: str) -> List[str]:
        soup = BeautifulSoup(content, 'html.parser')

        tweet_spans = soup.find_all('span', class_='css-1jxf684 r-bcqeeo r-1ttztb7 r-qvutc0 r-poiln3')

        tweets = []
        for span in tweet_spans:
            tweet_text = span.get_text(strip=True)
            if tweet_text:
                tweets.append(tweet_text)

        return tweets

    def _extract_tweets(self, limit: int = None) -> TwitterContent:
        tweet_elements = self.driver.find_elements(By.XPATH, "//div[@data-testid='tweet']")

        if limit:
            tweet_elements = tweet_elements[:limit]

        tweets = []
        for tweet in tweet_elements:
            try:
                text_element = tweet.find_element(By.XPATH, ".//div[@data-testid='tweetText']")
                time_element = tweet.find_element(By.XPATH, ".//time")
                tweets.append(Tweet(
                    text=text_element.text,
                    timestamp=time_element.get_attribute('datetime')
                ))
            except Exception as e:
                logger.error(f"Error extracting tweet: {e}")

        return TwitterContent(
            tweets=tweets,
            page_source=self.driver.page_source
        )

    def cleanup(self):
        if self.driver:
            self.driver.quit()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

def get_tweets(num_tweets=5):
    with TwitterScraper(headless=False) as scraper:
        scraper.login_to_twitter("skarandom","SK99@pass")
        return [x.text for x in scraper.get_home_page_content(num_tweets).tweets]

# Usage example
if __name__ == "__main__":
    print(get_tweets(2))


from .text_splitter import *
from .query_refiner import *
from .search_retriever import *
from .wordllama_embeds import WordLLamaRetriever
from .search_queries_generator import SearchQueryToNSubquery
from .query_refiner import QueryRefiner


import re
import sys
import json
import logging
import os

from langchain_core.messages import SystemMessage, HumanMessage

from .prompts import PROMPTS
from hugchat import hugchat
from hugchat.login import Login

os.environ['TF_ENABLE_ONEDNN_OPTS'] = '0'
from datetime import datetime
from PyQt5.QtWidgets import QApplication, QWidget, QLineEdit, QVBoxLayout, QTextEdit, QDesktopWidget, QHBoxLayout, \
    QComboBox
from PyQt5.QtCore import Qt, QPropertyAnimation, QEasingCurve, QRect, pyqtSignal, QObject
from PyQt5.QtGui import QPainter, QColor, QFont, QTextCursor

import threading
# Register QTextCursor for use in signals
from PyQt5.QtCore import QMetaType

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

QMetaType.type("QTextCursor")

DEFAULT_MODELS: list = ['Qwen/Qwen2.5-72B-Instruct', 'meta-llama/Meta-Llama-3.1-70B-Instruct',
                        'CohereForAI/c4ai-command-r-plus-08-2024', 'meta-llama/Llama-3.2-11B-Vision-Instruct',
                        'NousResearch/Hermes-3-Llama-3.1-8B', 'mistralai/Mistral-Nemo-Instruct-2407',
                        'microsoft/Phi-3.5-mini-instruct']


class StreamHandler(QObject):
    new_token_signal = pyqtSignal(str)

    def __init__(self, response_callback):
        super().__init__()
        self.response_callback = response_callback
        self.full_response = ""

    def on_llm_new_token(self, token: str, **kwargs):
        self.new_token_signal.emit(token)
        self.full_response += token

    def on_llm_end(self, response, **kwargs):
        self.response_callback(self.full_response)


class SpotlightLLM(QWidget):
    def __init__(self, email,password,models: list | str = None, execution_mode="Local"):
        super().__init__()
        self.email = email
        self.password = password  # NOTE: This should be kept securely. Use environment variables or secure vaults.
        self.models = models or DEFAULT_MODELS
        self.execution_mode = execution_mode  # Default to Local execution
        self.current_model = self.models[0]
        self.chatbot = self.initialize_chat()
        self.initUI()
        self.setup_data_directory()
        self.session_interactions = []
        self.response_complete = threading.Event()

        # Set up logging
        logging.basicConfig(filename='spotlight_llm.log', level=logging.DEBUG,
                            format='%(asctime)s - %(levelname)s - %(message)s')

    def initialize_chat(self):
        cookie_path_dir = "./cookies/"  # NOTE: trailing slash (/) is required to avoid errors
        sign = Login(self.email, self.password)
        cookies = sign.login(cookie_dir_path=cookie_path_dir, save_cookies=True)
        qwen_model = self.models[0] if isinstance(self.models, list) else self.models
        chatbot = hugchat.ChatBot(cookies=cookies.get_dict(),
                                  default_llm=qwen_model)
        return chatbot

    def initUI(self):
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)
        self.setAttribute(Qt.WA_TranslucentBackground)

        layout = QVBoxLayout()
        layout.setContentsMargins(20, 20, 20, 20)

        # Create a horizontal layout for search bar, model selection, and execution mode
        search_layout = QHBoxLayout()

        self.search_bar = QLineEdit(self)
        self.search_bar.setStyleSheet("""
            QLineEdit {
                background-color: rgba(60, 60, 60, 200);
                border: none;
                border-radius: 20px;
                padding: 10px;
                color: white;
                font-size: 18px;
            }
        """)
        self.search_bar.returnPressed.connect(self.on_submit)
        search_layout.addWidget(self.search_bar, 5)

        self.model_selector = QComboBox(self)
        self.model_selector.addItems(self.models)
        self.model_selector.setStyleSheet(self.get_combobox_style())
        self.model_selector.currentTextChanged.connect(self.on_model_change)
        search_layout.addWidget(self.model_selector, 2)

        self.execution_selector = QComboBox(self)
        self.execution_selector.addItems(["Local", "GPU"])
        self.execution_selector.setStyleSheet(self.get_combobox_style())
        self.execution_selector.currentTextChanged.connect(self.on_execution_mode_change)
        search_layout.addWidget(self.execution_selector, 2)

        layout.addLayout(search_layout)

        self.result_area = QTextEdit(self)
        self.result_area.setReadOnly(True)
        self.result_area.setStyleSheet("""
            QTextEdit {
                background-color: rgba(60, 60, 60, 200);
                border: none;
                border-radius: 10px;
                padding: 10px;
                color: white;
                font-size: 14px;
            }
        """)
        self.result_area.hide()
        layout.addWidget(self.result_area)

        self.setLayout(layout)

        screen = QDesktopWidget().screenNumber(QDesktopWidget().cursor().pos())
        screen_size = QDesktopWidget().screenGeometry(screen)
        window_width = 750
        window_height = 60
        x = (screen_size.width() - window_width) // 2
        y = screen_size.height() // 4
        self.setGeometry(x, y, window_width, window_height)

    def get_combobox_style(self):
        return """
            QComboBox {
                background-color: rgba(60, 60, 60, 200);
                border: none;
                border-radius: 20px;
                padding: 10px;
                color: white;
                font-size: 14px;
            }
            QComboBox::drop-down {
                border: none;
            }
            QComboBox::down-arrow {
                image: url(down_arrow.png);
                width: 12px;
                height: 12px;
            }
            QComboBox QAbstractItemView {
                background-color: rgba(60, 60, 60, 200);
                border: none;
                selection-background-color: rgba(80, 80, 80, 200);
                color: white;
            }
        """

    def setup_data_directory(self):
        self.data_dir = os.path.join(os.path.expanduser("~"), "spotlight_llm_data")
        os.makedirs(self.data_dir, exist_ok=True)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setBrush(QColor(30, 30, 30, 200))
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(self.rect(), 20, 20)

    def on_model_change(self, model):
        self.current_model = model
        logging.info(f"Model changed to: {model}")

    def on_execution_mode_change(self, mode):
        self.execution_mode = mode
        logging.info(f"Execution mode changed to: {mode}")

    def on_submit(self):
        query = self.search_bar.text()
        if query:
            self.result_area.clear()
            self.result_area.show()
            self.animate_expand()
            self.current_query = query
            self.current_timestamp = datetime.now().isoformat()
            self.response_complete.clear()
            threading.Thread(target=self.get_response, args=(query,), daemon=True).start()

    def update_result_area(self, token):
        cursor = self.result_area.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(token)
        self.result_area.setTextCursor(cursor)
        self.result_area.ensureCursorVisible()

    def add_universe_prompt(self, prompt):
        return PROMPTS.RAR_PROMPT + f"\n\n {prompt}"

    def add_messages_prompt(self, prompt):
        return [
            # {"role": "system","content": CLAUDE_SYSTEM_PROMPT},
            # SystemMessage(content=CLAUDE_SYSTEM_PROMPT),
            SystemMessage(content=PROMPTS.RAR_PROMPT),
            # {"role": "user","content": prompt}
            HumanMessage(content=prompt)
        ]

    def format_results_multi_agent(self, results):
        output = []

        # Add the final answer
        output.append("Final Answer:\n")
        output.append(results['final_answer'])
        output.append("")

        # Add subqueries and their results
        output.append("Subqueries and WebAnalysis:\n")
        for i, subquery in enumerate(results['subqueries']):
            output.append(f"{i + 1}. Subquery: {subquery}\n")

            # Get the corresponding result
            result = results['subquery_results'][i]

            # output.append("   Web Result:")
            # output.append(f"   {result['web_result']}")

            output.append("Relevant Information:")
            output.append(f"{result['relevant_info']}")
            output.append("")

        return "\n".join(output)

    def filter_prompts(self, prompt):
        # Add your prompt filtering logic here
        filtered_prompt = " ".join(prompt.split()[:-1])

        prompt = re.split("User Question : ", filtered_prompt)[-1]
        return prompt

    def get_response(self, user_query):
        # prompt = self.add_universe_prompt(user_query)
        prompt = user_query
        logging.debug(f"Starting response generation in {self.execution_mode} mode")
        stream_handler = StreamHandler(self.save_interaction)
        stream_handler.new_token_signal.connect(self.update_result_area)

        try:
            web_search_flag = False
            if prompt.lower().split()[-1] in ["google", "bing"]:
                prompt = self.filter_prompts(prompt)
                web_search_flag = True
            for resp in self.chatbot.chat(
                    prompt,
                    stream=True,
                    web_search=web_search_flag
            ):
                if resp:
                    stream_handler.on_llm_new_token(resp['token'])

            stream_handler.on_llm_end(None)
        except Exception as e:
            error_message = f"An error occurred: {str(e)}"
            QApplication.instance().postEvent(self.result_area, QTextCursor(self.result_area.document()))
            self.result_area.setPlainText(error_message)
            self.save_interaction(error_message)
            logging.exception(f"Error in {self.execution_mode} mode")
        finally:
            if self.execution_mode == "GPU":
                # Add any necessary cleanup for GPU resources
                pass
            logging.debug(f"Finished response generation in {self.execution_mode} mode")
            self.response_complete.set()

    def save_interaction(self, response):
        interaction_data = {
            "timestamp": self.current_timestamp,
            "model": self.current_model,
            "execution_mode": self.execution_mode,
            "query": self.current_query,
            "response": response
        }
        self.session_interactions.append(interaction_data)
        logging.info(f"Interaction saved: {self.current_timestamp}")

    def animate_expand(self):
        self.animation = QPropertyAnimation(self, b"geometry")
        self.animation.setDuration(300)
        self.animation.setStartValue(self.geometry())
        new_height = 400
        new_geometry = QRect(self.x(), self.y(), self.width(), new_height)
        self.animation.setEndValue(new_geometry)
        self.animation.setEasingCurve(QEasingCurve.OutCubic)
        self.animation.start()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.close()
    def closeEvent(self, event):
        logging.info("Application closing")
        super().closeEvent(event)

def spotlight_llm_run(email,password):
    app = QApplication(sys.argv)
    ex = SpotlightLLM(email=email,password=password)
    ex.show()
    sys.exit(app.exec_())

if __name__ == '__main__':
    spotlight_llm_run(email="santhoshkammari1999@gmail.com",password="SK99@pass123")


import string

from langchain_core.language_models import BaseChatModel
from nltk import sent_tokenize, FreqDist
import nltk
nltk.download("stopwords")
from nltk.corpus import stopwords

from ..query_master import SearchQueryToNSubquery,SearchRetriever,QueryRefiner

class AdvancedAISearcher:
    def __init__(self, chunk_overlap=25,
                 chunk_size=250,
                 max_urls=5,
                 n_key_sentences = 25,
                 topk = 10):
        self.n_key_sentences = n_key_sentences
        self.chunk_overlap = chunk_overlap
        self.chunk_size = chunk_size
        self.max_urls = max_urls
        self.topk = topk

    def generic_search(self, llm: BaseChatModel, query: str,
               verbose:bool = False,
               return_content_list = False,
                       n_splits = None)-> str:
        refined_query = QueryRefiner.refine_query(llm=llm, query=query)
        if verbose:
            print(f"Refined Query: {refined_query}")

        query_splits = SearchQueryToNSubquery.ai_splits(llm=llm, query=refined_query,
                                                        n_splits=n_splits)
        if verbose:
            print(f"Query Splits: {query_splits}")
        retriever = SearchRetriever(
            chunk_overlap=self.chunk_overlap,
            chunk_size=self.chunk_size,
            max_urls=self.max_urls,
        )
        ans = []
        all_urls = []
        for chunk in query_splits.get("refined_splits", []):
            results = retriever.query_based_content_retrieval(chunk,verbose=verbose,topk=self.topk)
            ans.extend(results.topk_chunks)
            all_urls.extend(results.urls)
        join_urls = "\n".join(all_urls)
        if return_content_list:
            return ans,all_urls
        ans = "\n".join(ans)
        final_ans = f"Answer:\n{ans}\n\nURLs:\n{join_urls}"
        return final_ans

    def search(self, llm: BaseChatModel, query: str,
               verbose:bool = False)-> str:
        refined_query = QueryRefiner.refine_query(llm=llm, query=query)
        if verbose:
            print(f"Refined Query: {refined_query}")

        query_splits = SearchQueryToNSubquery.ai_splits(llm=llm, query=refined_query)
        if verbose:
            print(f"Query Splits: {query_splits}")
        retriever = SearchRetriever(
            chunk_overlap=self.chunk_overlap,
            chunk_size=self.chunk_size,
            max_urls=self.max_urls,
        )
        ans = []
        all_urls = []
        for chunk in query_splits.get("refined_splits", []):
            results = retriever.query_based_content_retrieval(chunk,verbose=verbose)
            ans.extend(results.topk_chunks)
            all_urls.extend(results.urls)
        join_urls = "\n".join(all_urls)

        all_sentences = []
        for i, answer in enumerate(ans):
            all_sentences.extend(self.__preprocess_text(answer, i))
        key_sentences = self.__extract_key_sentences(all_sentences, n=self.n_key_sentences)

        context = "\n".join(list(set([i[1] for i in key_sentences])))
        final_ans = f"Answer:\n{context}\n\nURLs:\n{join_urls}"
        return final_ans

    def __preprocess_text(self,text, i):
        # Tokenize the text into sentences
        org_sentences = sent_tokenize(text)

        # Remove punctuation and convert to lowercase
        sentences = [s.translate(str.maketrans('', '', string.punctuation)).lower() for s in org_sentences]

        # Remove stopwords
        stop_words = set(stopwords.words('english'))
        sentences = [' '.join([word for word in s.split() if word not in stop_words]) for s in sentences]
        sentences = [(i, os, s) for os, s in zip(org_sentences, sentences)]
        return sentences

    def __extract_key_sentences(self,sentences, n=3):
        # Combine all sentences
        all_words = ' '.join([_[1] for _ in sentences]).split()

        # Calculate word frequencies
        freq_dist = FreqDist(all_words)

        # Score sentences based on word frequencies
        sentence_scores = []
        for idx, org_sent, sentence in sentences:
            score = sum(freq_dist[word] for word in sentence.split())
            sentence_scores.append((idx, org_sent, sentence, score))

        # Sort sentences by score and return top n
        return [s for s in sorted(sentence_scores, key=lambda x: x[-1], reverse=True)[:n]]

    def generate_final_answer(self,llm, query,verbose = False,n_splits = None):
        # Preprocess and extract key information
        answers, urls = self.generic_search(llm, query, return_content_list=True,
                                   verbose=verbose,
                                            n_splits = n_splits)
        all_sentences = []
        for i, answer in enumerate(answers):
            all_sentences.extend(self.__preprocess_text(answer, i))
        key_sentences = self.__extract_key_sentences(all_sentences, n=self.n_key_sentences)

        context = "\n".join(list(set([i[1] for i in key_sentences])))
        prompt = f"""Based on the following key information:

    {context}

    Please provide a brief answer to the query: "{query}"
    Focus only on the most crucial points."""

        response = llm.invoke(prompt).content
        return response



# from ..semantic_chunker import SemanticSimilarityTextChunker
#
#
# class ContextExtractor:
#     def __init__(self, top_chunks_in_article=3, sentence_chunk_size_in_article=3, max_subquery_relevant_chunks=3):
#         self.top_chunks_in_article = top_chunks_in_article
#         self.sentence_chunk_size_in_article = sentence_chunk_size_in_article
#         self.max_subquery_relevant_chunks = max_subquery_relevant_chunks
#
#     def extract_context(self, query, articles):
#         chunker = SemanticSimilarityTextChunker(query)
#         all_chunks = []
#
#         for article in articles:
#             chunks = chunker.chunk_text(
#                 article,
#                 top_chunks_in_article=self.top_chunks_in_article,
#                 chunk_size=self.sentence_chunk_size_in_article
#             )
#             all_chunks.extend(chunks)
#
#         all_chunks.sort(
#             key=lambda chunk: chunker.score_sentence(chunk, chunker.get_query_keywords()),
#             reverse=True
#         )
#
#         relevant_chunks = all_chunks[:self.max_subquery_relevant_chunks]
#         return relevant_chunks
#
# # Example usage:
# # extractor = ContextExtractor(top_chunks_in_article=5, sentence_chunk_size_in_article=3, max_subquery_relevant_chunks=10)
# # relevant_chunks = extractor.extract_context("example query", ["article1 text", "article2 text"])


# from .main.main import MultiAgentQueryOrchestrator
# from .main import ContextExtractor

# from .web_search import WebSearchAgent,FastHTMLParserV3,OptimizedMultiQuerySearcher,WebSearchArguments
from .main.search_provider.searcher import RealTimeGoogleSearchProvider
from .main.parse_url.html_parser import UrlTextParser


import asyncio
import re
from typing import List

from .base import Searcher


class GoogleSearch(Searcher):
    def perform_search(self, query: str) -> List[str]:
        # Here you'd implement the actual API search using google_searcher
        print(f"Searching Google for: {query}")
        return [f"Result {i + 1} for {query}" for i in range(5)]  # Dummy results

    def perform_batch_search(self, batch_queries: List[str]) -> List[str]:
        pass

import asyncio
from typing import List
from ....web_search import OptimizedMultiQuerySearcher

class RealTimeGoogleSearchProvider(Searcher):
    def __init__(self, search_provider="google",chromedriver_path="/usr/local/bin/chromedriver", max_workers=None,
                 animation = False):
        self.search_provider = search_provider
        self.chromedriver_path = chromedriver_path
        self.max_workers = max_workers
        self.animation = animation


    def perform_search(self, query: str,max_urls=5) -> List[str]:
        with OptimizedMultiQuerySearcher(chromedriver_path=self.chromedriver_path,
                                                    max_workers=self.max_workers,
                                                    animation=self.animation) as searcher:
            return searcher.search_single_query(query,search_provider=self.search_provider).urls[:max_urls]


    async def _async_batch_search(self, batch_queries,max_urls=5) -> List[str]:
        with OptimizedMultiQuerySearcher(chromedriver_path=self.chromedriver_path,
                                         max_workers=self.max_workers,
                                         animation=self.animation) as searcher:
            all_urls = await searcher.search_multiple_queries(
                queries=batch_queries,
                search_provider=self.search_provider
            )
            all_urls = [url.urls for url in all_urls]
            filtered_urls = [y for x in zip(*all_urls) for y in x]
            filtered_urls = [self.extract_until_hash(x) if self.is_hash(x) else x for x in filtered_urls]
            filtered_urls = [_ for _ in filtered_urls if _]
            return filtered_urls[:max_urls]

    def perform_batch_search(self, batch_queries,max_urls=5) -> List[str]:
        return asyncio.run(self._async_batch_search(batch_queries,max_urls=max_urls))

    def is_hash(self, x):
        return '#' in x

    def extract_until_hash(self, x):
        results = re.findall(r'(.*)#',x)
        if results:
            return results[0]
        return ""


import os
import platform
import subprocess
import sys
import urllib.request
import zipfile
import shutil

import nltk


def is_windows():
    return platform.system().lower() == "windows"


def is_chrome_installed():
    if is_windows():
        return shutil.which("chrome.exe") is not None
    else:
        return shutil.which("google-chrome") is not None


def get_chrome_version():
    try:
        if is_windows():
            result = subprocess.run(
                ["reg", "query", "HKEY_CURRENT_USER\\Software\\Google\\Chrome\\BLBeacon", "/v", "version"],
                capture_output=True, text=True, check=True)
            return result.stdout.strip().split()[-1]
        else:
            result = subprocess.run(["google-chrome", "--version"], capture_output=True, text=True, check=True)
            return result.stdout.strip().split()[-1]
    except subprocess.CalledProcessError:
        return None


def install_chrome():
    if is_windows():
        chrome_url = "https://dl.google.com/chrome/install/latest/chrome_installer.exe"
        chrome_installer = "chrome_installer.exe"
        urllib.request.urlretrieve(chrome_url, chrome_installer)
        subprocess.run([chrome_installer, "/silent", "/install"], check=True)
        os.remove(chrome_installer)
    else:
        chrome_url = "https://dl.google.com/linux/direct/google-chrome-stable_current_amd64.deb"
        chrome_file = "google-chrome-stable_current_amd64.deb"
        urllib.request.urlretrieve(chrome_url, chrome_file)
        subprocess.run(["sudo", "dpkg", "-i", chrome_file], check=True)
        subprocess.run(["sudo", "apt-get", "install", "-f", "-y"], check=True)
        os.remove(chrome_file)

def setup_nltk():
    try:
        nltk.data.find('tokenizers/punkt')
        nltk.data.find('corpora/stopwords')
    except LookupError:
        print("Downloading necessary NLTK data...")
        nltk.download('punkt')
        nltk.download('stopwords')

def setup_local():
    setup_nltk()
    if not is_chrome_installed():
        print("Google Chrome is not installed. Installing...")
        install_chrome()

    chrome_version = get_chrome_version()
    if not chrome_version:
        print("Failed to detect Chrome version. Please install Chrome manually.")
        return

    print(f"Detected Chrome version: {chrome_version}")
    major_version = chrome_version.split('.')[0]

    # Download ChromeDriver
    if is_windows():
        chromedriver_url = f"https://edgedl.me.gvt1.com/edgedl/chrome/chrome-for-testing/{chrome_version}/win64/chromedriver-win64.zip"
        chromedriver_zip = "chromedriver-win64.zip"
    else:
        chromedriver_url = f"https://edgedl.me.gvt1.com/edgedl/chrome/chrome-for-testing/{chrome_version}/linux64/chromedriver-linux64.zip"
        chromedriver_zip = "chromedriver-linux64.zip"

    print(f"Downloading ChromeDriver from: {chromedriver_url}")
    urllib.request.urlretrieve(chromedriver_url, chromedriver_zip)

    # Extract ChromeDriver
    with zipfile.ZipFile(chromedriver_zip, 'r') as zip_ref:
        zip_ref.extractall()

    # Move ChromeDriver to appropriate location
    if is_windows():
        chromedriver_path = os.path.join("chromedriver-win64", "chromedriver.exe")
        target_path = "C:\\Windows\\System32\\chromedriver.exe"
    else:
        chromedriver_path = os.path.join("chromedriver-linux64", "chromedriver")
        target_path = "/usr/local/bin/chromedriver"

    if is_windows():
        shutil.move(chromedriver_path, target_path)
    else:
        os.chmod(chromedriver_path, 0o755)
        subprocess.run(["sudo", "mv", chromedriver_path, target_path], check=True)

    # Clean up downloaded files
    os.remove(chromedriver_zip)
    if is_windows():
        shutil.rmtree("chromedriver-win64")
    else:
        shutil.rmtree("chromedriver-linux64")

    # Print versions
    print("\nChromeDriver version:")
    subprocess.run(["chromedriver", "--version"], check=True)
    print("\nGoogle Chrome version:")
    if is_windows():
        subprocess.run(["chrome.exe", "--version"], check=True)
    else:
        subprocess.run(["google-chrome", "--version"], check=True)


if __name__ == "__main__":
    if not is_windows() and os.geteuid() != 0:
        print("This script requires root privileges on Linux. Please run with sudo.")
        sys.exit(1)
    setup_local()


from langchain_ollama import ChatOllama
from .ai_searcher import AdvancedAISearcher

from .query_master import SearchRetriever

from langchain_huggy import HuggyLLM
# query  = "explain me Design of ACM , better if we use arxiv paper"
def ai_search(query,verbose: bool|str= False,
              llm = None,
              chunk_overlap=25,
              chunk_size=250,
              max_urls=5,
              n_key_sentences=25,
              topk=10,
              n_web_queries_to_generate = 5
              ):
    rtr = AdvancedAISearcher(
        chunk_overlap=chunk_overlap,
        chunk_size=chunk_size,
        max_urls=max_urls,
        n_key_sentences=n_key_sentences,
        topk=topk
    )
    if isinstance(verbose,str): verbose = verbose.lower() == 'true'
    res = rtr.generate_final_answer(llm, query,verbose=verbose,n_splits=n_web_queries_to_generate)
    return res

def search(query,verbose: bool|str = False,
           llm= None):
    rtr = AdvancedAISearcher()
    if isinstance(verbose,str): verbose = verbose.lower() == 'true'
    res = rtr.search(llm,query,verbose=verbose)
    return res

def fast_search(query,verbose: bool|str = False,
                llm= None):
    rtr = AdvancedAISearcher(chunk_overlap=20,
                             chunk_size=100,
                             max_urls=2,
                             n_key_sentences=10)
    if isinstance(verbose,str): verbose = verbose.lower() == 'true'
    res = rtr.search(llm,query,verbose=verbose)
    return res

def google_search(query,verbose: bool|str = False):
    retriever = SearchRetriever(
        chunk_overlap=20,
        chunk_size=100,
        max_urls=2,
    )
    if isinstance(verbose,str): verbose = verbose.lower() == 'true'
    results = retriever.query_based_content_retrieval(query, topk=5,verbose=verbose)
    return "\n".join(results.topk_chunks)

def deep_google_search(query,verbose: bool|str = False):
    retriever = SearchRetriever()
    if isinstance(verbose,str): verbose = verbose.lower() == 'true'
    results = retriever.query_based_content_retrieval(query, topk=5,verbose=verbose)
    return "\n".join(results.topk_chunks)




import re
from typing import List
import concurrent.futures
import torch
from wordllama import WordLlama

class WordLLamaRetriever:
    def __init__(self, text, batch_size=32):
        self.wl = WordLlama.load()
        self.text = text
        self.embeds = None
        self.sents = None
        self.batch_size = batch_size
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.__build_embeds()

    @staticmethod
    def _split_sentences(text):
        return re.findall(r'[^!.?]+[!.?]', text) if isinstance(text, str) else text

    def embed_batch(self, sents_batch):
        with torch.no_grad():
            return torch.tensor(self.wl.embed(sents_batch)).to(self.device)

    def get_embeds(self, text) -> tuple:
        sents: List = self._split_sentences(text)
        embeds = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            for i in range(0, len(sents), self.batch_size):
                sents_batch = sents[i:i + self.batch_size]
                futures.append(executor.submit(self.embed_batch, sents_batch))

            for future in concurrent.futures.as_completed(futures):
                embeds.append(future.result())

        if embeds:
            embeds = torch.cat(embeds, dim=0)
        else:
            embeds = None

        return embeds, sents

    def top_k(self, query, k=5):
        q_embeds, _ = self.get_embeds(query)

        if self.embeds is None or q_embeds is None:
            return []

        # Normalize query embeddings
        q_embeds = torch.nn.functional.normalize(q_embeds, p=2, dim=1)

        # Compute cosine similarity
        similarity = torch.mm(self.embeds, q_embeds.t())

        # Get top k similarities and indices
        top_k_similarities, top_k_indices = torch.topk(similarity.squeeze(), k)

        top_sents = [self.sents[i] for i in top_k_indices]
        return top_sents

    def __build_embeds(self):
        self.embeds, self.sents = self.get_embeds(self.text)
        if self.embeds is not None:
            # Normalize embeddings
            self.embeds = torch.nn.functional.normalize(self.embeds, p=2, dim=1)

if __name__ == '__main__':
    import time

    with open("/home/ntlpt59/Downloads/combined_text.txt", "r") as file:
        large_text = file.read()

    start_time = time.time()
    ht = WordLLamaRetriever(large_text)
    build_time = time.time() - start_time

    query_time = time.time()
    sents = ht.top_k("what is date of coo?")
    query_time = time.time() - query_time

    print(f"Build time: {build_time:.2f} seconds")
    print(f"Query time: {query_time:.2f} seconds")



import PyPDF2
from wordllama import WordLlama
import tiktoken
import re


class PDFContentExtractor:
    def __init__(self, pdf_path, chunk_size=1000):
        self.pdf_path = pdf_path
        self.chunk_size = chunk_size
        self.wl = WordLlama.load()
        self.enc = tiktoken.get_encoding("cl100k_base")
        self.content = self._extract_pdf_content()
        self.chunks = self._create_chunks()

    def _extract_pdf_content(self):
        with open(self.pdf_path, 'rb') as file:
            reader = PyPDF2.PdfReader(file)
            content = ""
            for page in reader.pages:
                content += page.extract_text()
        return re.sub(r'\s+', ' ', content).strip()

    def _create_chunks(self):
        tokens = self.enc.encode(self.content)
        chunks = []
        for i in range(0, len(tokens), self.chunk_size):
            chunk = self.enc.decode(tokens[i:i + self.chunk_size])
            chunks.append(chunk)
        return chunks

    def get_relevant_content(self, query, top_k=5):
        ranked_chunks = self.wl.rank(query, self.chunks)
        return [chunk for chunk, score in ranked_chunks[:top_k]]

    def search(self, query, top_k=5):
        relevant_chunks = self.get_relevant_content(query, top_k)
        for i, chunk in enumerate(relevant_chunks, 1):
            print(f"\nChunk {i}:")
            print(chunk)
            print("-" * 50)


# Example usage
if __name__ == "__main__":
    pdf_path = "path/to/your/pdf/file.pdf"
    extractor = PDFContentExtractor(pdf_path)

    query = "Example query about a specific topic"
    extractor.search(query)